#ifndef WakeComH
#define WakeComH

#include <SysUtils.hpp>
#include <SyncObjs.hpp>
#include <Classes.hpp>

#define DEFAULT_BAUD_RATE  19200 //default baudrate
#define DEFAULT_TX_TIMEOUT    50 //default TX timeout, ms
#define DEFAULT_RX_TIMEOUT    50 //default RX timeout, ms
#define LONG_RX_TIMEOUT     2000 //long RX timeout, ms

#define DEFAULT_RTS_STATE  RTS_CONTROL_TOGGLE  //default RTS state
#define DEFAULT_DTR_STATE  DTR_CONTROL_DISABLE //default DTR state

#define FRAME     250   //Wake frame size

//Error codes:
enum WakeErr_t
{
  ERR_NO, //no error
  ERR_TX, //Rx/Tx error
  ERR_BU, //device busy error
  ERR_RE, //device not ready error
  ERR_PA, //parameters value error
  ERR_NR, //no replay
  ERR_NC, //no carrier
  ERR_EE  //EEPROM error
};

//Standard commands codes:
enum WakeCmd_t
{
  CMD_NOP,     //no operation
  CMD_ERR,     //frame receive error
  CMD_ECHO,    //echo
  CMD_INFO,    //read device info
  CMD_SETADDR, //set address
  CMD_GETADDR  //get address
};

#define LO(x)    (x & 0xFF)
#define HI(x)    ((x >> 8) & 0xFF)
#define BYTE1(x) LO(x)
#define BYTE2(x) HI(x)
#define BYTE3(x) ((x >> 16) & 0xFF)
#define BYTE4(x) ((x >> 24) & 0xFF)
#define WORD(h, l) ((h << 8) | l)
#define DWORD(b3,b2,b1,b0) ((WORD(b3,b2) << 16) | WORD(b1,b0))

#define _ENTER_COMMAND(x) x->Enter();try{
#define _LEAVE_COMMAND(x) }__finally{x->Leave();}

//----------------- WAKE protocol implementation class: ---------------------

class TWakeCom : public TObject
{
private:
  HANDLE hCom;         //port handle
  DCB dcb;             //DCB structure
  DCB dcbc;            //saved DCB structure
  COMMTIMEOUTS ComTo;  //timeouts structure
  COMMTIMEOUTS ComToc; //saved timeouts structure
  AnsiString FName;    //logical port name
  int FNumber;         //port number
  int FBaudRate;       //baud rate
  bool Platform_NT;    //platform type flag
  WORD BaseAddress;    //UART base address (not for NT systems)
  bool RtsToggle;      //RTS toggle flag (not for NT systems)

  void __fastcall SetNumber(int number);     //set port number
  void __fastcall SetName(AnsiString name);  //set port name

  enum WStuff_t
  {
    FEND  = 0xC0,        //Frame END
    FESC  = 0xDB,        //Frame ESCape
    TFEND = 0xDC,        //Transposed Frame END
    TFESC = 0xDD         //Transposed Frame ESCape
  };
  enum WCrc_t
  {
    CRC_INIT = 0xDE      //initial CRC value
  };
  int FTxAddr;           //transmitted address
  int FTxCmd;            //transmitted command code
  int FTxPtr;            //transmit buffer pointer
  int FRxPtr;            //receive buffer pointer
  unsigned char RxN;     //received bytes number
  AnsiString FTxCmdName; //transmitted command name
  AnsiString FDeviceName; //device name
  void __fastcall RxByte(unsigned char &b); //RX byte
  BYTE __fastcall PortIn(WORD Address); //port read (not for NT systems)
  void __fastcall ByteStuff(unsigned char b, DWORD &bptr, unsigned char *buff);
  void __fastcall ByteDeStuff(unsigned char &b);
  void __fastcall DowCRC(unsigned char b, unsigned char &crc); //CRC calculation
  void __fastcall RxFrame(unsigned char &Addr,
                          unsigned char &Cmd,
                          unsigned char &Nbt,
                          unsigned char *Data); //RX frame
  void __fastcall TxFrame(unsigned char Addr,
                          unsigned char Cmd,
                          unsigned char Nbt,
                          unsigned char *Data); //TX frame
  unsigned char FTxData[FRAME]; //TX buffer
  unsigned char FRxData[FRAME]; //RX buffer

protected:
  TCriticalSection *CmdCrSec; //critical section for commands execution
  void __fastcall NewCmd(int cmd, AnsiString name); //new frame create
  void __fastcall AddByte(unsigned char b);  //add byte to buffer
  void __fastcall AddWord(unsigned short b); //add word to buffer
  void __fastcall AddDWord(unsigned int b);  //add double word to buffer
  void __fastcall AddData(char *d, int count); //add data to buffer
  void __fastcall SendCmd(int a = 0); //TX frame to the given address
  unsigned char __fastcall  GetByte(void);   //read byte from buffer
  unsigned short __fastcall  GetWord(void);  //read word from buffer
  unsigned int __fastcall  GetDWord(void);   //read double word from buffer
  void __fastcall GetData(char *d, int count); //read data from buffer
  int __fastcall GetRxN(void); //read received bytes number

public:
  __fastcall TWakeCom();  //constructor
  __fastcall ~TWakeCom(); //destructor

  int __fastcall NameToNumber(AnsiString name); //port name to number
  AnsiString __fastcall NumberToName(int number); //port number to name
  void __fastcall GetPortList(TStringList *APortBox); //get port list
  bool __fastcall Access(int PortNumber); //check port accessibility by number
  bool __fastcall Access(AnsiString PortName); //check port accessibility by name
  void __fastcall Open(int PortNumber); //open port by number
  void __fastcall Open(AnsiString PortName); //open port by name
  void __fastcall GetDeviceList(TStringList *ADeviceBox); //get device list
  void __fastcall GetNetDeviceList(AnsiString PortName,      //get network
                                   TStringList *ADeviceBox); //device list
  void __fastcall OpenDevice(AnsiString DevName); //open port by device name
  void __fastcall Close(); //close port
  void __fastcall SetBaudRate(int baudrate); //set baud rate
  void __fastcall SetMode(DWORD RtsMode, DWORD DtrMode); //set mode
  void __fastcall SetTo(DWORD RxTo, DWORD TxTo); //set timeouts
  void __fastcall Purge(void); //port buffer clear

  __property int Number = {read=FNumber, write=SetNumber}; //port number
  __property AnsiString PortName = {read=FName, write=SetName}; //port name
  __property int BaudRate = {read=FBaudRate, write=SetBaudRate}; //baudrate
  __property AnsiString DeviceName = {read=FDeviceName}; //device name

  AnsiString __fastcall GetInfo(int a = 0); //read device info
};

//---------------------------------------------------------------------------

#endif
